/*
 * Copyright (C) 2007-2009 KenD00
 * 
 * This file is part of DumpHD.
 * 
 * DumpHD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package dumphd.util;

/**
 * Class used for returning size and crc32 from a copied / decrypted file.
 * 
 * @author KenD00
 */
public class CopyResult {

   public long size = 0;
   public long crc32 = 0;

   public CopyResult() {
      // Nothing
   }

   public CopyResult(long length, long crc32) {
      this.size = length;
      this.crc32 = crc32;
   }

}
