package bdvm.vm;

/**
 * General purpose exception for the BDVM.
 * 
 * @author KenD00
 */
public class BDVMException extends Exception {
   
   /**
    * The SerialVersionUID of this class, automatically generated, just here to avoid stupid warnings.
    * Should be replaced by some better defined value if serialization is used someday.
    */
   private static final long serialVersionUID = -2292581931925100157L;

   public BDVMException() {
      super();
   }
   
   public BDVMException(String message) {
      super(message);
   }
   
   public BDVMException(String message, Throwable cause) {
      super(message, cause);
   }
   
   public BDVMException(Throwable cause) {
      super(cause);
   }
}
